/*
 * canfield !! 
 *
 * $Header: can.c,v 0.8 94/10/31 15:17:14 miyazawa Rel $
 */

#include "modoki.h"

#define _MAX 52
#define _MIN 36
#define TOP_X 120
#define TOP_Y 10

static int card[_MAX];
static int found[4];
static int from;
static int reach;
static int px[11], py[11];
static int holdcard;
static int kk;
static struct tableau *tableau1[4];
static struct tableau *tableau2[4];
static struct tableau *stock, *sc;
static int whenHold();
static int whenNotHold();

void
initCan()
{
	static int check();
	static void restart();

	_ckhook = check;
	_rshook = restart;
	_when_hold = whenHold;
	_when_not_hold = whenNotHold;
	current = CANFIELD;
	(*_rshook)();
	(*_ckhook)();
}

static void
restart()
{
	register i;
	struct tableau *temp;
	static long seed;

	srandom(randomseed);
	XUnmapSubwindows(dpy, Table);
	clearStack();

	for (i = 0; i < 4; i++) {
		tfree(&tableau2[i]);
	}
	tfree(&stock);

	for (i = 0; i < 4; i++) {
		tableau2[i] = NULL;
	}

	/* 0 draw */
	/* 1 stock */
	/* 2 talon */
	/* 3-6 foundation */
	/* 7-10 tableau */

	px[0] = TOP_X;
	py[0] = TOP_Y;
	px[1] = TOP_X + 50;
	py[1] = TOP_Y;
	px[2] = TOP_X;
	py[2] = TOP_Y + 70;

	for (i = 0; i < 4; i++) {
		px[i + 3] = px[i + 7] = TOP_X + 50 + i * 50;
		py[i + 3] = TOP_Y + 140;
		py[i + 7] = TOP_Y + 70;
	}
	PUTCARD(FACE, 0);
	for (i = 1; i < 11; i++) {
		PUTCARD(FRAME, i);
	}
	XUnmapWindow(dpy, _decks[2 + JOKER].self);
	for (i = 0; i < 4; i++) {
		tfree(&tableau2[i]);
	}

	for (i = 0; i < 4; i++) {
		found[i] = -2;
	}
	for (i = 0; i < _MAX; i++) {
		card[i] = i;
	}
	for (i = 0; i < _MAX; i++) {
		swap(&card[random() % _MAX], &card[i]);
	}
	for (i = 0; i < _MIN; i++) {
		temp = talloc();
		temp->next = stock;
		temp->n = card[i];
		temp->tmp = 0;
		stock = temp;
	}
	kk = _MIN;
	for (i = 0; i < 4; i++) {
		temp = talloc();
		temp->next = NULL;
		temp->n = card[kk++];
		tableau1[i] = tableau2[i] = temp;
		PUTCARD(temp->n, i + 3);
	}
	PUTCARD(card[kk], 7);
	found[0] = card[kk];
	reach = NUMB(card[kk] + 12);
	PUTCARD(card[++kk], 2);

	holdcard = -1;
	havecard = False;
	from = 0;
	sc = NULL;
	_status = 0;
}

static int
check()
{
	register i, j;
	int a1, b1;
	struct tableau *temp;
	int sou, des;

	autopilot = False;
	sou = des = -1;
	if (stock == NULL || stock->next == NULL) {
		for (i = 0; i < 4; i++) {
			if (tableau1[i] != NULL &&
				tableau1[i]->next != NULL) {
				goto skip;
			}
		}
		autopilot = True;
		for (i = 0; i < 4; i++) {
			if (NUMB(found[i]) != reach) {
				goto skip;
			}
		}
		autopilot = False;
		return (1);
	}
skip:
	/* tableau & foundation */
	for (i = 0; i < 4; i++) {
		if (tableau2[i] != NULL) {
			b1 = tableau2[i]->n;
			for (j = 0; j < 4; j++) {
				a1 = found[j];
				if (NUMBP(a1 + 1, b1) && SUITP(a1, b1) ||
					a1 < 0 && NUMBP(b1, reach + 1)) {
					des = j + 7;
					sou = i + 3;
					goto ok;
				}
			}
		}
		/* both tableau */
		if (tableau1[i] != NULL) {
			b1 = tableau1[i]->n;
			for (j = 0; j < 4; j++) {
				if (i == j) {
					continue;
				}
				if (tableau2[j] != NULL) {
					a1 = tableau2[j]->n;
					if (NUMBP(b1 + 1, a1) &&
						!COLORP(a1, b1) &&
						!NUMBP(a1, reach + 1)) {
						des = j + 3;
						sou = i + 3;
						goto ok;
					}
				}
			}
		}
	}
	if (kk < _MAX) {
		b1 = card[kk];
		/* pile & foundation */
		for (i = 0; i < 4; i++) {
			a1 = found[i];
			if (NUMBP(a1 + 1, b1) && SUITP(a1, b1) ||
			    a1 < 0 && NUMBP(b1, reach + 1)) {
				des = i + 7;
				sou = 2;
				goto ok;
			}
		}
		/* pile & tableau */
		for (i = 0; i < 4; i++) {
			if (tableau2[i] != NULL) {
				a1 = tableau2[i]->n;
				if (NUMBP(b1 + 1, a1) && !COLORP(a1, b1)) {
					des = i + 3;
					sou = 2;
					goto ok;
				}
			}
			else {
				des = i + 3;
				sou = 2;
				goto ok;
			}
		}
	}
	if (sc != NULL) {
		b1 = sc->n;
		/* stock & foundation */
		for (i = 0; i < 4; i++) {
			a1 = found[i];
			if (NUMBP(a1 + 1, b1) && SUITP(a1, b1) ||
			    a1 < 0 && NUMBP(b1, reach + 1)) {
				des = i + 7;
				sou = 1;
				goto ok;
			}
		}
		/* stock & tableau */
		for (i = 0; i < 4; i++) {
			if (tableau2[i] != NULL) {
				a1 = tableau2[i]->n;
				if (NUMBP(b1 + 1, a1) && !COLORP(a1, b1)) {
					des = i + 3;
					sou = 1;
					goto ok;
				}
			}
			else if (kk == _MAX) {
				des = i + 3;
				sou = 1;
				goto ok;
			}
		}
		/* stock read in advance */
		if (sc->next == NULL) {
			temp = stock;
			while (temp != NULL) {
				if (temp->next != NULL) {
					temp = temp->next;
				}
				if (temp->next != NULL) {
					temp = temp->next;
				}
				b1 = temp->n;
				/* foundation */
				for (i = 0; i < 4; i++) {
					a1 = found[i];
					if (NUMBP(a1 + 1, b1) &&
						SUITP(a1, b1) ||
						a1 < 0 &&
						NUMBP(b1, reach + 1)) {
						sou = des = 0;
						goto ok;
					}
				}
				/* tableau */
				for (i = 0; i < 4; i++) {
					if (tableau2[i] != NULL) {
						a1 = tableau2[i]->n;
						if (NUMBP(b1 + 1, a1) &&
							!COLORP(a1, b1)) {
							sou = des = 0;
							goto ok;
						}
					}
					else if (kk == _MAX) {
						sou = des = 0;
						goto ok;
					}
				}
				if (temp->next != NULL) {
					temp = temp->next;
				}
				else {
					temp = NULL;
				}
			}
			return (-1);
		}
	}
	if (stock == NULL && sou < 0) {
		return (-1);
	}
	sou = des = 0;
ok:
	if (autopilot) {
		if (havecard) {
			jumpMove(px[des], py[des], holdcard);
			whenHold(des);
			havecard = False;
		}
		else {
			whenNotHold (sou);
		}
	}
	return (0);
}

static int
whenNotHold(pos)
int pos;
{
	struct tableau *tmp;
	register i;

	if (pos <= 6 && pos >= 3) {
		if (tableau2[pos - 3] != NULL) {
			holdcard = tableau2[pos - 3]->n;
			havecard = True;
			from = pos;
			return (0);
		}
	}
	else if (pos == 2) {
		if (kk < _MAX) {
			holdcard = card[kk];
			havecard = True;
			from = pos;
			return (0);
		}
	}
	else if (pos == 1) {
		if (sc != NULL) {
			holdcard = sc->n;
			havecard = True;
			from = pos;
			return (0);
		}
	}
	else if (pos == 0) {
		if (sc == NULL) {
			if (stock != NULL) {
				sc = stock;
				for (i = 0; i < 2; i++) {
					if (sc->next != NULL) {
						sc = sc->next;
					}
				}
				PUTCARD(sc->n, 1);
				if (sc->next == NULL) {
					PUTCARD(FRAME, 0);
				}
				return (0);
			}
		}
		else {
			if (sc->next == NULL) {
				sc = NULL;
				PUTCARD(FRAME, 1);
				PUTCARD(FACE, 0);
			}
			else {
				for (i = 0; i < 3; i++) {
					if (sc->next != NULL) {
						sc = sc->next;
					}
				}
				PUTCARD(sc->n, 1);
				if (sc->next == NULL) {
					PUTCARD(FRAME, 0);
				}
			}
			return (0);
		}
	}
	return (-1);
}

static int
whenHold(pos)
int pos;
{
	struct tableau *tmp0, *tmp1;
	struct tableau *temp;
	register i;
	int dif;
	static void delstock();

	if (pos == from || pos < 0) {
		holdcard = -1;
		return (-1);
	}
	if (pos >= 7) {	/* to Foundation */
		if (NUMBP(found[pos - 7] + 1, holdcard) &&
			SUITP(found[pos - 7], holdcard) ||
			found[pos - 7] < 0 &&
			NUMBP(holdcard, reach + 1)) {
			found[pos - 7] = holdcard;
			PUTCARD(holdcard, pos);
			if (from == 1) {
				delstock();
			}
			else if (from == 2) {
				kk++;
				if (kk < _MAX) {
					PUTCARD(card[kk], from);
				}
			}
			else {
				tmp1 = tableau2[from - 3];
				if (tmp1->next == NULL) {
					tableau1[from - 3] =
					tableau2[from - 3] = NULL;
				}
				else {
					py[from] -= 20;
					PUTCARD(tmp1->next->n, from);
					tableau2[from - 3] = tmp1->next;
				}
				XtFree((XtPointer) tmp1);
			}
			holdcard = -1;
			return (0);
		}
	}
	else if (pos <= 6 && pos >= 3) {	/* to Tableau */
		if (from <= 2) {
			tmp1 = tableau2[pos - 3];
			if ((tmp1 != NULL) && NUMBP(holdcard + 1, tmp1->n) &&
				!COLORP(holdcard, tmp1->n) &&
				!NUMBP(tmp1->n, reach + 1) ||
				tmp1 == NULL && (from == 2 || kk == _MAX)) {
				temp = talloc();
				temp->next = tmp1;
				temp->n = holdcard;
				tableau2[pos - 3] = temp;
				if (tmp1 != NULL) {
					py[pos] += 20;
				}
				else {
					tableau1[pos - 3] = temp;
				}
				PUTCARD(holdcard, pos);
				if (from == 1) {
					delstock();
				}
				else {
					kk++;
					if (kk < _MAX) {
						PUTCARD(card[kk], from);
					}
				}
				holdcard = -1;
				return (0);
			}
		}
		else {
			tmp0 = tableau1[from - 3];
			tmp1 = tableau2[pos - 3];
			if (tmp1 != NULL && NUMBP(tmp0->n + 1, tmp1->n) &&
				!COLORP(tmp0->n, tmp1->n) &&
				!NUMBP(tmp1->n, reach + 1)) {
				tmp1 = tableau2[from - 3];
				dif = NUMB(tmp0->n) - NUMB(tmp1->n) + 1;
				if (dif < 0) {
					dif = 13 + dif;
				}
				py[from] = TOP_Y + 140;
				tableau1[from - 3] = tableau2[from - 3] = NULL;
				temp = tmp1;
				for (i = 0; i < dif; i++) {
					pushCard(temp);
					temp = temp->next;
				}
				while (temp = popCard()) {
					py[pos] += 20;
					PUTCARD(temp->n, pos);
				}
				tmp0->next = tableau2[pos - 3];
				tableau2[pos - 3] = tmp1;
				holdcard = -1;
				return (0);
			}
		}
	}
	holdcard = -1;
	return (-1);
}

static void
delstock()
{
	struct tableau *temp;
	
	if (stock == sc) {
		temp = sc;
		stock = temp->next;
		sc = NULL;
		XtFree((XtPointer) temp);
	}
	else {
		for (temp = stock; temp->next != sc; temp = temp->next);
		temp->next = sc->next;
		XtFree((XtPointer) sc);
		sc = temp;
		PUTCARD(sc->n, 1);
		sc->tmp = 1;
	}
}
