/*
 * corners !! 
 *
 * $Header: cnr.c,v 0.6 94/10/31 15:17:18 miyazawa Rel $
 */

#include "modoki.h"

#define _MAX 52
#define TOP_X 60
#define TOP_Y 60

static int card[_MAX];
static int found[4];
static int px[11], py[11];
static int holdcard;
static int from;
static int sp;
static int reach;
static struct tableau *tableau[5], *temp;
static struct tableau *pile;
static int whenHold();
static int whenNotHold();

void
initCnr()
{
	static int check();
	static void restart();

	_ckhook = check;
	_rshook = restart;
	_when_hold = whenHold;
	_when_not_hold = whenNotHold;
	current = CORNERS;
	(*_rshook)();
	(*_ckhook)();
}

static void
restart()
{
	register i;
	static long seed;

	srandom(randomseed);
	XUnmapSubwindows(dpy, Table);
	clearStack();

	px[0] = TOP_X;
	py[0] = TOP_Y;

	px[1] = TOP_X + 80;
	py[1] = TOP_Y;

	px[2] = TOP_X + 80;
	py[2] = TOP_Y + 80;
	px[3] = TOP_X;
	py[3] = TOP_Y + 160;
	px[4] = TOP_X + 80;
	py[4] = TOP_Y + 160;
	px[5] = TOP_X + 160;
	py[5] = TOP_Y + 160;
	px[6] = TOP_X + 80;
	py[6] = TOP_Y + 240;

	px[7] = TOP_X;
	py[7] = TOP_Y + 80;
	px[8] = TOP_X + 160;
	py[8] = TOP_Y + 80;
	px[9] = TOP_X;
	py[9] = TOP_Y + 240;
	px[10] = TOP_X + 160;
	py[10] = TOP_Y + 240;

	for (i = 0; i < 5; i++) {
		tfree(&tableau[i]);
	}
	tfree(&pile);

	for (i = 0; i < _MAX; i++) {
		card[i] = i;
	}
	for (i = 0; i < _MAX; i++) {
		swap(&card[random() % _MAX], &card[i]);
	}

	reach = NUMB(card[0] + 12);
	found[0] = card[0];
	sp = 1;
	for (i = 0; i < 5; i++) {
		tableau[i] = talloc();
		tableau[i]->n = card[sp++];
		tableau[i]->next = NULL;
	}

	PUTCARD(card[sp], 0);
	PUTCARD(FRAME, 1);
	for (i = 2; i < 7; i++) {
		PUTCARD(FRAME, i);
		PUTCARD(tableau[i - 2]->n, i);
	}
	PUTCARD(found[0], 7);
	for (i = 8; i < 11; i++) {
		found[i - 7] = -2;
		PUTCARD(FRAME, i);
	}

	holdcard = -1;
	havecard = False;
	from = 0;
	_status = 0;
}

static int
check()
{
	register i, j, k;
	int sou, des;

	autopilot = False;
	sou = des = -1;
	for (i = 0; i < 4; i++) {
		if (NUMB(found[i]) != reach) {
			goto skip;
		}
	}
	return (1);
skip:

	/* stock to foundation */
	if (sp < _MAX) {
		for (i = 0; i < 4; i++) {
			if ((found[i] < 0 && NUMB(reach + 1)
			     == NUMB(card[sp])) ||
			    (NUMB(found[i] + 1) == NUMB(card[sp])
			     && SUIT(found[i]) == SUIT(card[sp]))) {
				sou = 0;
				des = i + 7;
				goto ok;
			}
		}
	}
	/* pile to foundation */
	if (pile != NULL) {
		for (i = 0; i < 4; i++) {
			if ((found[i] < 0 && NUMB(reach + 1)
			     == NUMB(pile->n)) ||
			    (NUMB(found[i] + 1) == NUMB(pile->n)
			     && SUIT(found[i]) == SUIT(pile->n))) {
				sou = 1;
				des = i + 7;
				goto ok;
			}
		}
	}
	/* tableau to foundation */
	for (i = 0; i < 5; i++) {
		if (tableau[i] != NULL) {
			for (j = 0; j < 4; j++) {
				if (((found[j] < 0 && NUMB(reach + 1)
				      == NUMB(tableau[i]->n)) ||
				     (NUMB(found[j] + 1) == NUMB(tableau[i]->n)
				      && SUIT(found[j]) == SUIT(tableau[i]->n)
				      ))) {
					sou = i + 2;
					des = j + 7;
					goto ok;
				}
			}
		}
	}
	/* check empty tableau */
	for (i = 0; i < 5; i++) {
		if (tableau[i] == NULL) {
			sou = i + 2;
			goto ok;
		}
	}
	for (i = 0; i < 5; i++) {
		char used[5];
		/* stock to tableau */
		if (sp < _MAX && NUMB(tableau[i]->n) == NUMB(card[sp] + 1)) {
			sou = 0;
			des = i + 2;
			goto ok;
		}
		/* pile to tableau */
		if (pile != NULL && NUMB(tableau[i]->n) == NUMB(pile->n + 1)) {
			sou = 1;
			des = i + 2;
			goto ok;
		}
		/* tableau to tableau */
		temp = tableau[i];
		for (j = 0; j < 5; j++) {
			used[j] = 0;
		}
		used[i] = 1;
retry:
		for (j = 0; j < 5; j++) {
			if (!used[j] && NUMB(temp->n + 1)
			    == NUMB(tableau[j]->n)) {
				used[j] = 1;
				temp = temp->next;
				if (temp == NULL) {
					sou = i + 2;
					goto ok;
				}
				for (k = 0; k < 4; k++) {
					if (((found[k] < 0 && NUMB(reach + 1)
					      == NUMB(temp->n)) ||
					     (NUMB(found[k] + 1)
					      == NUMB(temp->n) &&
					      SUIT(found[k])
					      == SUIT(temp->n)))) {
						sou = i + 2;
						goto ok;
					}
				}
				goto retry;
			}
		}
	}
	if (sp == _MAX) {
		return (-1);
	}
	ok:
	if (sp == _MAX && found[1] >= 0 && found[2] >= 0 && found[3] >= 0 &&
	    (pile == NULL || pile != NULL && pile->next == NULL)) {
		autopilot = True;
		if (havecard) {
			jumpMove(px[des], py[des], holdcard);
			whenHold(des);
			havecard = False;
		}
		else {
			whenNotHold (sou);
		}
	}
	return (0);
}

static int
whenNotHold(pos)
int pos;
{
	switch (pos) {
	case 0:
		if (sp < _MAX) {
			holdcard = card[sp];
			havecard = True;
			from = 0;
			return (0);
		}
		break;
	case 1:
		if (pile != NULL) {
			holdcard = pile->n;
			havecard = True;
			from = 1;
			return (0);
		}
		break;
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
		if (tableau[pos - 2] != NULL) {
			holdcard = tableau[pos - 2]->n;
			havecard = True;
			from = pos;
			return (0);
		}
		break;
	default:
		break;
	}
	return (-1);
}

static int
whenHold(pos)
int pos;
{
	struct tableau *temp;
	Boolean f;

	if (pos == from || pos < 0) {
		holdcard = -1;
		return (-1);
	}
	switch (pos) {
	case 1:
		if (from == 0) {
			if (pile != NULL) {
				px[1] += 8;
			}
			temp = talloc();
			temp->next = pile;
			temp->n = holdcard;
			pile = temp;
			PUTCARD(holdcard, pos);
			if (sp == _MAX - 1) {
				sp = _MAX;
			}
			else if (sp < _MAX)
				PUTCARD(card[++sp], 0);
			holdcard = -1;
			return (0);
		}
		break;
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
		if (tableau[pos - 2] == NULL ||
		    NUMB(tableau[pos - 2]->n) == NUMB(holdcard + 1)) {
			PUTCARD(holdcard, pos);
			switch (from) {
			case 0:
				temp = talloc();
				temp->next = tableau[pos - 2];
				temp->n = holdcard;
				tableau[pos - 2] = temp;
				if (sp == _MAX - 1) {
					sp = _MAX;
				}
				else if (sp < _MAX) {
					PUTCARD(card[++sp], 0);
				}
				break;
			case 1:
				temp = pile;
				pile = pile->next;
				temp->next = tableau[pos - 2];
				tableau[pos - 2] = temp;
				if (pile != NULL) {
					px[1] -= 8;
					PUTCARD(pile->n, 1);
				}
				break;
			case 2:
			case 3:
			case 4:
			case 5:
			case 6:
				temp = tableau[from - 2];
				tableau[from - 2] = tableau[from - 2]->next;
				temp->next = tableau[pos - 2];
				tableau[pos - 2] = temp;
				if (tableau[from - 2] != NULL) {
					PUTCARD(tableau[from - 2]->n, from);
				}
				PUTCARD(holdcard, pos);
			}
			holdcard = -1;
			return (0);
		}
		break;
	case 7:
	case 8:
	case 9:
	case 10:
		f = False;
		if (found[pos - 7] < 0) {
			if (NUMB(reach + 1) == NUMB(holdcard)) {
				f = True;
			}
		}
		else if (NUMB(found[pos - 7] + 1) == NUMB(holdcard)
			 && SUIT(found[pos - 7]) == SUIT(holdcard)) {
			if (NUMB(found[pos - 7]) == NUMB(reach + 1)) {
				px[pos] += 8 * (pos % 2 ? 1 : -1);
				py[pos] += 8 * (pos < 9 ? 1 : -1);
			}
			f = True;
		}
		if (f) {
			found[pos - 7] = holdcard;
			switch (from) {
			case 0:
				if (sp == _MAX - 1) {
					sp = _MAX;
				}
				else if (sp < _MAX) {
					PUTCARD(card[++sp], 0);
				}
				break;
			case 1:
				temp = pile->next;
				XtFree((XtPointer) pile);
				pile = temp;
				if (pile != NULL) {
					px[1] -= 8;
					PUTCARD(pile->n, 1);
				}
				break;
			case 2:
			case 3:
			case 4:
			case 5:
			case 6:
				temp = tableau[from - 2]->next;
				XtFree((XtPointer) tableau[from - 2]);
				tableau[from - 2] = temp;
				if (tableau[from - 2] != NULL) {
					PUTCARD(temp->n, from);
				}
				break;
			}
			PUTCARD(holdcard, pos);
			holdcard = -1;
			return (0);
		}
		break;
	}
	holdcard = -1;
	return (-1);
}
