/*
 * klondike !! 
 *
 * $Header: kdk.c,v 0.8 94/10/31 15:17:20 miyazawa Rel $
 */

#include "modoki.h"

#define _MAX 52
#define _MIN 24
#define TOP_X 10
#define TOP_Y 20

static int card[_MAX];
static int found[4];
static int from;
static int holdcard;
static int px[13], py[13];
static struct tableau *tableau1[7];
static struct tableau *tableau2[7];
static struct tableau *stock, *sc;
static int whenHold();
static int whenNotHold();

void
initKdk()
{
	static int check();
	static void restart();

	_ckhook = check;
	_rshook = restart;
	_when_hold = whenHold;
	_when_not_hold = whenNotHold;
	current = KLONDIKE;
	(*_rshook)();
	(*_ckhook)();
}

static void
restart()
{
	register i, j, k;
	struct tableau *temp;
	static long seed;

	srandom(randomseed);
	XUnmapSubwindows(dpy, Table);
	clearStack();

	px[0] = TOP_X;
	py[0] = TOP_Y;
	px[1] = TOP_X + 60;
	py[1] = TOP_Y;

	for (i = 0; i < 7; i++) {
		px[i + 2] = TOP_X + 60 + i * 60;
		py[i + 2] = TOP_Y + 80;
	}
	for (i = 0; i < 4; i++) {
		px[i + 9] = TOP_X + 180 + i * 80;
		py[i + 9] = TOP_Y;
	}

	for (i = 0; i < 7; i++) {
		tfree(&tableau2[i]);
	}
	tfree(&stock);

	for (i = 0; i < 4; i++) {
		PUTCARD(FRAME, i + 9);
		found[i] = -1;
	}

	for (i = 0; i < _MAX; i++) {
		card[i] = i;
	}

	for (i = 0; i < _MAX; i++) {
		swap(&card[random() % _MAX], &card[i]);
	}

	for (i = 0; i < _MIN; i++) {
		temp = talloc();
		temp->next = stock;
		temp->n = card[i];
		stock = temp;
	}
	PUTCARD(FRAME, 1);

	PUTCARD(FACE, 0);
	for (i = 0; i < 4; i++){
		PUTCARD(FRAME, i + 1);
	}

	k = _MIN;
	for (i = 0; i < 7; i++) {
		PUTCARD(FRAME, i + 2);
		for (j = 0; j < i + 1; j++) {
			temp = talloc();
			temp->next = tableau2[i];
			temp->n = card[k++];
			tableau1[i] = tableau2[i] = temp;
			if (j == i) {
				PUTCARD(temp->n, i + 2);
			}
		}
	}

	sc = NULL;
	holdcard = -1;
	havecard = False;
	from = 0;
	_status = 0;
}

static int
check()
{
	register i, j;
	register a1, b1;
	int sou, des;
	struct tableau *temp;

	autopilot = False;
	sou = des = 0;
	if (stock == NULL || stock->next == NULL) {
		for (i = 0; i < 7; i++) {
			if (tableau1[i] != NULL && tableau1[i]->next != NULL) {
				goto skip;
			}
		}
		autopilot = True;
		for (i = 0; i < 4; i++) {
			if (NUMB(found[i]) != KING) {
				goto skip;
			}
		}
		autopilot = False;
		return (1);
	}
skip:
	/* tableau & foundation */
	for (i = 0; i < 7; i++) {
		if (tableau2[i] != NULL) {
			b1 = tableau2[i]->n;
			for (j = 0; j < 4; j++) {
				a1 = found[j];
				if (BINDP(a1, b1) || a1 < 0 &&
					NUMB(b1) == ACE) {
					des = j + 9;
					sou = i + 2;
					goto ok;
				}
			}

		}
		/* both tableau */
		if (tableau1[i] != NULL) {
			b1 = tableau1[i]->n;
			for (j = 0; j < 7; j++) {
				if (i == j) {
					continue;
				}
				if (tableau2[j] != NULL) {
					a1 = tableau2[j]->n;
					if (STRAIGHTP(b1, a1) &&
						!COLORP(a1, b1)) {
						des = j + 2;
						sou = i + 2;
						goto ok;
					}
				}
				else {
					if (NUMB(b1) == KING &&
						tableau1[i]->next != NULL) {
						des = j + 2;
						sou = i + 2;
						goto ok;
					}
				}
			}
		}
	}
	
	if (sc != NULL) {
		b1 = sc->n;
		/* stock & foundation */
		for (i = 0; i < 4; i++) {
			a1 = found[i];
			if (BINDP(a1, b1) || a1 < 0 && NUMB(b1) == ACE) {
				des = i + 9;
				sou = 1;
				goto ok;
			}
		}
		for (i = 0; i < 7; i++) {
			if (tableau2[i] != NULL) {
				a1 = tableau2[i]->n;
				if (STRAIGHTP(b1, a1) && !COLORP(a1, b1)) {
					des = i + 2;
					sou = 1;
					goto ok;
				}
			}
			else if (NUMB(b1) == KING) {
				des = i + 2;
				sou = 1;
				goto ok;
			}
		}
		if (sc->next == NULL) {
			temp = stock;
			while (temp != NULL) {
				if (temp->next != NULL) {
					temp = temp->next;
				}
				if (temp->next != NULL) {
					temp = temp->next;
				}
				b1 = temp->n;
				for (i = 0; i < 4; i++) {
					a1 = found[i];
					if (BINDP(a1, b1) || a1 < 0 &&
						NUMB(b1) == ACE) {
						goto ok;
					}
				}
				for (i = 0; i < 7; i++) {
					if (tableau2[i] != NULL) {
						a1 = tableau2[i]->n;
						if (STRAIGHTP(b1, a1) &&
							!COLORP(a1, b1)) {
							goto ok;
						}
					}
					else if (NUMB(b1) == KING) {
						goto ok;
					}
				}
				if (temp->next != NULL) {
					temp = temp->next;
				}
				else {
					temp = NULL;
				}
			}
			autopilot = False;
			return (-1);
		}
	}
	if (stock == NULL && sou == 0) {
		autopilot = False;
		return (-1);
	}
ok:
	if (autopilot) {
		if (havecard) {
			jumpMove(px[des], py[des], holdcard);
			whenHold(des);
			havecard = False;
		}
		else {
			whenNotHold (sou);
		}
	}
	return (0);
}

static int
whenNotHold(pos)
int pos;
{
	register i;

	if (pos < 9 && pos > 1) {
		if (tableau2[pos - 2] != NULL) {
			holdcard = tableau2[pos - 2]->n;
			havecard = True;
			from = pos;
			return (0);
		}
	}
	else if (pos == 1) {
		if (sc != NULL) {
			holdcard = sc->n;
			havecard = True;
			from = pos;
			return (0);
		}
	}
	else if (pos == 0) {
		if (sc == NULL) {
			if (stock != NULL) {
				sc = stock;
				for (i = 0; i < 2; i++) {
					if (sc->next != NULL) {
						sc = sc->next;
					}
				}
				PUTCARD(sc->n, 1);
				if (sc->next == NULL) {
					PUTCARD(FRAME, 0);
				}
				return (0);
			}
		}
		else {
			if (sc->next == NULL) {
				sc = NULL;
				PUTCARD(FRAME, 1);
				PUTCARD(FACE, 0);
			}
			else {
				for (i = 0; i < 3; i++) {
					if (sc->next != NULL) {
						sc = sc->next;
					}
				}
				PUTCARD(sc->n, 1);
				if (sc->next == NULL) {
					PUTCARD(FRAME, 0);
				}
			}
			return (0);
		}
	}
	return (-1);
}

static int
whenHold(pos)
int pos;
{
	register i;
	int dif;
	struct tableau *tmp0, *tmp1;
	struct tableau *temp;
	static void delstock();

	if (pos == from || pos < 0) {
		holdcard = -1;
		return (-1);
	}
	if (pos >= 9) {	/* to Foundation */
		if ((NUMB(holdcard) == ACE && found[pos - 9] < 0) ||
		    (SUIT(holdcard) == SUIT(found[pos - 9]) &&
		  found[pos - 9] + 1 == holdcard)) {
			found[pos - 9] = holdcard;
			PUTCARD(holdcard, pos);
			if (from == 1) {
				delstock();
			}
			else {
				tmp1 = tableau2[from - 2];
				if (tmp1->next == NULL) {
					py[from] = TOP_Y + 80;
				}
				else {
					if (py[from] > TOP_Y + 80) {
						py[from] -= 16;
					}
					PUTCARD(tmp1->next->n, from);
				}
				if (tmp1 == tableau1[from - 2]) {
					tableau1[from - 2] = tmp1->next;
				}
				tableau2[from - 2] = tmp1->next;
				XtFree((XtPointer) tmp1);
			}
			holdcard = -1;
			return (0);
		}
	}
	else if (pos > 1 && pos < 9) {	/* to Tableau */
		if (from == 1) {
			tmp1 = tableau2[pos - 2];
			if (tmp1 != NULL && NUMBP(holdcard + 1, tmp1->n) &&
			!COLORP(holdcard, tmp1->n) ||
			    NUMB(holdcard) == KING && tmp1 == NULL) {
				delstock();
				temp = talloc();
				temp->next = tmp1;
				temp->n = holdcard;
				tableau2[pos - 2] = temp;
				if (tmp1 != NULL) {
					py[pos] += 16;
				}
				else {
					tableau1[pos - 2] = temp;
				}
				PUTCARD(holdcard, pos);
				holdcard = -1;
				return (0);
			}
		}
		else {
			tmp0 = tableau1[from - 2];
			tmp1 = tableau2[pos - 2];
			if (tmp1 != NULL && NUMBP(tmp0->n + 1, tmp1->n) &&
			    !COLORP(tmp0->n, tmp1->n) ||
			    NUMB(tmp0->n) == KING && tmp1 == NULL) {
				py[from] = TOP_Y + 80;
				if (tmp1 == NULL) {
					PUTCARD(tmp0->n, pos);
				}
				if (tmp0->next != NULL) {
					PUTCARD(tmp0->next->n, from);
				}
				tmp0 = tableau1[from - 2];
				tmp1 = tableau2[from - 2];
				dif = NUMB(tmp0->n) - NUMB(tmp1->n) + 1;
				if (tableau2[pos - 2] == NULL) {
					dif--;
				}
				temp = tmp1;
				for (i = 0; i < dif; i++) {
					pushCard(temp);
					temp = temp->next;
				}
				while (temp = popCard()) {
					py[pos] += 16;
					PUTCARD(temp->n, pos);
				}
				tableau1[from - 2] =
					tableau2[from - 2] = tmp0->next;
				tmp0->next = tableau2[pos - 2];
				tableau2[pos - 2] = tmp1;
				if (tableau1[pos - 2] == NULL) {
					tableau1[pos - 2] = tmp0;
					tableau1[pos - 2]->next = NULL;
				}
				holdcard = -1;
				return (0);
			}
		}
	}
	holdcard = -1;
	return (-1);
}

static void
delstock()
{
	struct tableau *temp;

	if (stock == sc) {
		temp = sc;
		stock = temp->next;
		sc = NULL;
		XtFree((XtPointer) temp);
	}
	else {
		for (temp = stock; temp->next != sc; temp = temp->next);
		temp->next = sc->next;
		XtFree((XtPointer) sc);
		sc = temp;
		putCard(px[1], py[1], sc->n, 1);
	}
}
