/*
 * $Header: main.c,v 0.13 94/10/31 15:17:21 miyazawa Rel $
 *
 * Copyright (c) 1992  Advanced Integration Technology, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of AIT not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  AIT makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * Author:  Kunio Miyazawa
 *
 */

#include <stdio.h>
#include <locale.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/X10.h>
#include <X11/bitmaps/gray3>
#include <X11/cursorfont.h>
#include "modoki.h"

#include "card.bm"
#include "icon.bm"

#define CMAX 80
#define JSTEP 10
#define def_width 500
#define def_height 500

static XAssocTable *Key;

static Pixel Red, Green, White, Black;
static GC ngc, rgc, bgc, frame_gc, red_gc;
static Pixmap pixmaps[VARIETY];

static Boolean iscolor = False;
static Boolean nodrag;
static Boolean rvflag;
static Boolean lgflag;
static int jstep;
char *solitaire[GAMEMAX];

void (*_rshook)();
int (*_ckhook)();
Boolean havecard;
Boolean autopilot;
long randomseed;
String current;
String gamefile;
int _status;

Window Table;
struct card _decks[CMAX];
Display *dpy;
int scr;

static void setIcon();
static void mainLoop();
static void initTableWindow();
static void initCardWindows();

#ifdef DEBUG
static int (*oldhook)();

static int
errorHook(d, e)
Display *d;
XErrorEvent *e;
{
	gamefile = "/tmp/xsols.debug";
	(void) saveData();
	(*oldhook)(d, e);
}
#endif


main(ac, av)
int ac;
char **av;
{
	static char filename[1024];
	static XrmOptionDescRec options[] = {
		{"-reverse", "*reverse", XrmoptionNoArg, "True"},
		{"-nodrag", "*noDrag", XrmoptionNoArg, "True"},
		{"-loadgame", "*loadGame", XrmoptionNoArg, "True"},
	};
	static XtResource resources[] = {
		{"reverse", "Reverse", XtRBoolean, sizeof(Boolean),
		(Cardinal) &rvflag, XtRImmediate, (caddr_t) False},
		{"noDrag", "NoDrag", XtRBoolean, sizeof(Boolean),
		(Cardinal) &nodrag, XtRImmediate, (caddr_t) False},
		{"jumpStep", "JumpStep", XtRInt, sizeof(int),
		(Cardinal) &jstep, XtRImmediate, (caddr_t) JSTEP},
		{"blackColor", "Foreground", XtRPixel, sizeof(Pixel),
		(Cardinal) &Black, XtRString, "black"},
		{"whiteColor", "Background", XtRPixel, sizeof(Pixel),
		(Cardinal) &White, XtRString, "white"},
		{"redColor", "Foreground", XtRPixel, sizeof(Pixel),
		(Cardinal) &Red, XtRString, "red"},
		{"greenColor", "Background", XtRPixel, sizeof(Pixel),
		(Cardinal) &Green, XtRString, "green"},
		{"gameFile", "GameFile", XtRString, sizeof(String),
		(Cardinal) &gamefile, XtRString, ""},
		{"loadGame", "LoadGame", XtRBoolean, sizeof(Boolean),
		(Cardinal) &lgflag, XtRImmediate, (caddr_t) False},
		{"game1", "Game1", XtRString, sizeof(String),
		(Cardinal) &solitaire[0], XtRString, CANFIELD},
		{"game2", "Game2", XtRString, sizeof(String),
		(Cardinal) &solitaire[1], XtRString, PYRAMID},
		{"game3", "Game3", XtRString, sizeof(String),
		(Cardinal) &solitaire[2], XtRString, TOWERS},
		{"game4", "Game4", XtRString, sizeof(String),
		(Cardinal) &solitaire[3], XtRString, CORNERS},
		{"game5", "Game5", XtRString, sizeof(String),
		(Cardinal) &solitaire[4], XtRString, QUADRILLE},
		{"game6", "Game6", XtRString, sizeof(String),
		(Cardinal) &solitaire[5], XtRString, GOLF},
		{"game7", "Game7", XtRString, sizeof(String),
		(Cardinal) &solitaire[6], XtRString, CALCULATION},
		{"game8", "Game8", XtRString, sizeof(String),
		(Cardinal) &solitaire[7], XtRString, KLONDIKE},
		{"game9", "Game9", XtRString, sizeof(String),
		(Cardinal) &solitaire[8], XtRString, SHUFFLE},
	};
	static Arg my_args[] = {
		{XtNwidth, (XtArgVal) def_width},
		{XtNheight, (XtArgVal) def_height},
	};
	XtAppContext app;
	Widget top, w;
	Widget createBoard();
	void (*getGame())(), (*first_game)();
	extern String fallbacks[];

	setlocale(LC_ALL,"");
	top = XtAppInitialize(&app, "XSols",
		  options, XtNumber(options), &ac, av, fallbacks, NULL, 0);
	if (ac > 1) {
		fprintf(stderr,
			"Usage: %s [-nodrag] [-reverse] [-loadgame]\n", av[0]);
		exit(1);
	}
	XtGetApplicationResources(top, 0, resources,
		XtNumber(resources), NULL, (Cardinal) 0);
	dpy = XtDisplay(top);
	scr = DefaultScreen(dpy);
	if (DisplayCells(dpy, scr) > 2) {
		iscolor = True;
	}
	if (jstep <= 0) {
		jstep = JSTEP;
	}
	if (strlen(gamefile) == 0) {
		char *pw = (char*) getenv("HOME");
		sprintf(filename, "%s/.xsols", pw ? pw : "/tmp");
		gamefile = (String) filename;
	}

	XtSetMappedWhenManaged(top, False);
	w = createBoard(top);
	XtSetValues(w, my_args, XtNumber(my_args));
	XtRealizeWidget(top);
	initTableWindow(w);
	setIcon(top);
	XtSetMappedWhenManaged(top, True);
	XtMapWidget(top);
	randomseed = time(0);
	first_game = getGame(solitaire[0]);
	first_game();
	if (lgflag) {
		loadData();
	}
#ifdef DEBUG
	oldhook = XSetErrorHandler(errorHook);
#endif
	mainLoop(app);
}

static void
setIcon(w)
Widget w;
{
	XWMHints hints;

	hints.flags = IconPixmapHint;
	hints.icon_pixmap = XCreateBitmapFromData
		(dpy, Table, icon_bits, icon_width, icon_height);
	XSetWMHints(dpy, XtWindow(w), &hints);
}

static void
initTableWindow(w)
Widget w;
{
	int i;
	static Arg arg = {XtNbackground, 0};

	ngc = XCreateGC(dpy, RootWindow(dpy, scr), 0, NULL);
	rgc = XCreateGC(dpy, RootWindow(dpy, scr), 0, NULL);
	bgc = XCreateGC(dpy, RootWindow(dpy, scr), 0, NULL);
	red_gc = XCreateGC(dpy, RootWindow(dpy, scr), 0, NULL);
	frame_gc = XCreateGC(dpy, RootWindow(dpy, scr), 0, NULL);

	XtGetValues(w, &arg, 1);
	XSetForeground(dpy, bgc, Black);
	XSetBackground(dpy, bgc, (Pixel) arg.value);

	XSetForeground(dpy, ngc, Black);
	XSetBackground(dpy, ngc, White);

	XSetForeground(dpy, rgc, White);
	XSetBackground(dpy, rgc, Black);

	XSetForeground(dpy, frame_gc, 1);
	XSetFunction(dpy, frame_gc, GXxor);
	XSetSubwindowMode(dpy, frame_gc, IncludeInferiors);

	XSetForeground(dpy, red_gc, Red);
	XSetBackground(dpy, red_gc, White);

	Table = XtWindow(w);

	XDefineCursor(dpy, Table, XCreateFontCursor(dpy, XC_hand2));

	initCardWindows(Table);

	XSelectInput(dpy, Table,
		ExposureMask | ButtonPressMask | PointerMotionMask);
}

static void
initCardWindows(win)
Window win;
{
	register int i;
	XAssocTable *XCreateAssocTable();
	XSetWindowAttributes watt;
	static struct card tako;
	static Pixmap makeCard();

	for (i = 0; i < VARIETY; i++) {
		pixmaps[i] = makeCard(i);
	}
	for (i = 0; i < CMAX; i++) {
		_decks[i].n = i < JOKER ? i : FRAME;
		_decks[i].self = XCreateSimpleWindow(dpy, win,
		       0, 0, card_width, card_height, 0, 0, White);
		XSelectInput(dpy, _decks[i].self,
			ButtonPressMask |
			((nodrag) ? 0 :
			(ButtonReleaseMask | OwnerGrabButtonMask)));
		XSetWindowBackgroundPixmap(dpy, _decks[i].self,
					   pixmaps[_decks[i].n]);
	}
	watt.backing_store = WhenMapped;
	if (DoesBackingStore(ScreenOfDisplay(dpy, scr))) {
		for (i = 0; i < CMAX; i++) {
			XChangeWindowAttributes(dpy, _decks[i].self,
				CWBackingStore, &watt);
		}
		XChangeWindowAttributes(dpy, Table, CWBackingStore, &watt);
	}

	Key = XCreateAssocTable(16);

	for (i = 0; i < CMAX; i++) {
		XMakeAssoc(dpy, Key, _decks[i].self, &_decks[i]);
	}
	XMakeAssoc(dpy, Key, win, &tako);
	tako.self = Table;
	tako.pos = -1;
}

static void
mainLoop(app)
XtAppContext app;
{
	XEvent e;
	struct card *where;
	extern void bluePopup();
	static void myDispatchEvent();
	int status;

	for (;;) {
		where = NULL;
		if (autopilot) {
			if (XtAppPending(app)) {
				XtAppNextEvent(app, &e);
				if (where = (struct card *) XLookUpAssoc
				    (dpy, Key, e.xany.window)) {
					if (e.xany.window == Table) {
						XtDispatchEvent(&e);
					}
				}
				else {
					XtDispatchEvent(&e);
				}
			}
		}
		else {
			XtAppNextEvent(app, &e);
			if (where = (struct card *) XLookUpAssoc
				(dpy, Key, e.xany.window)) {
				if (e.xany.window == Table) {
					XtDispatchEvent(&e);
				}
				myDispatchEvent(&e, where);
			}
			else {
				XtDispatchEvent(&e);
			}
		}
		if (!_status && (autopilot ||
			(where && ((struct card *) where)->pos >= 0))) {
			_status = (*_ckhook)();
			bluePopup(_status);
		}
	}
}

static void
myDispatchEvent(event, p)
XEvent *event;
struct card *p;
{
	static int oldx = -1, oldy;
	static int mx, my;
	static void drawFrame();
	static void holdCard();
	static void releaseCard();
	extern Boolean nodrag;

	switch (event->type) {
	case ButtonPress:
		if (event->xbutton.button != 1) {
			break;
		}
		if (havecard) {
			drawFrame(oldx, oldy, 0);
			releaseCard(p->pos, p->n);
			oldx = -1;
			havecard = False;
		}
		else {
			holdCard(p->pos, p->n);
			mx = event->xbutton.x;
			my = event->xbutton.y;
			drawFrame(oldx, oldy, 0);
		}
		break;
	case MotionNotify:
		if (event->xmotion.state != 256 && !nodrag) {
			break;
		}
		if (havecard) {
			drawFrame(oldx, oldy, 0);
			oldx = ((XMotionEvent *) event)->x_root - mx;
			oldy = ((XMotionEvent *) event)->y_root - my;
			drawFrame(oldx, oldy, 0);
		}
		break;
	case ButtonRelease:
		if (event->xbutton.button != 1) {
			break;
		}
		if (havecard) {
			drawFrame(oldx, oldy, 0);
			releaseCard(p->pos, p->n);
			oldx = -1;
			havecard = False;
		}
		break;
	}
}

static void
drawFrame(x, y, n)
int x, y, n;
{
	if (x == -1) {
		return;
	}
	XDrawRectangle(dpy, RootWindow(dpy, scr), frame_gc, x, y,
		card_width, card_height + n * 25);
}

void
putCard(x, y, n, p)
int x, y;
int n;
int p;
{
	if (n < JOKER) {
		_decks[n].pos = p;
		XMoveWindow(dpy, _decks[n].self, x, y);
		XMapRaised(dpy, _decks[n].self);
	}
	else {
		XSetWindowBackgroundPixmap
			(dpy, _decks[JOKER + p].self, pixmaps[n]);
		XClearWindow(dpy, _decks[JOKER + p].self);
		_decks[JOKER + p].pos = p;
		XMoveWindow(dpy, _decks[JOKER + p].self, x, y);
		XMapRaised(dpy, _decks[JOKER + p].self);
	}
}

void
jumpMove(x, y, n)
int x, y;
int n;
{
	XWindowAttributes wa;
	register x1, y1, x2, y2;
	register i, s, a;
	register dx, dy, step, r;

#define XSTEP 1
#define YSTEP XSTEP

	XGetWindowAttributes(dpy, _decks[n].self, &wa);
	XRaiseWindow(dpy, _decks[n].self);
	XSelectInput(dpy, Table, ButtonPressMask | PointerMotionMask);
	x2 = x / XSTEP;
	y2 = y / YSTEP;
	x1 = wa.x / XSTEP;
	y1 = wa.y / YSTEP;
	dx = x1 < x2 ? x2 - x1 : x1 - x2;
	dy = y1 < y2 ? y2 - y1 : y1 - y2;
	a = 1;
	if (dx > dy) {
		r = x1 < x2 ? 1 : -1;
		step = y1 < y2 ? 1 : -1;
		s = dx / 2;
		for (i = x1; i != x1 + dx * r + r; i += r) {
			if (a % jstep == 0) {
				XMoveWindow(dpy, _decks[n].self,
					    i * XSTEP, y1 * XSTEP);
			}
			s -= dy;
			if (s < 0) {
				s += dx;
				y1 += step;
			}
			a++;
		}
	}
	else {
		r = y1 < y2 ? 1 : -1;
		step = x1 < x2 ? 1 : -1;
		s = dy / 2;
		for (i = y1; i != y1 + dy * r + r; i += r) {
			if (a % jstep == 0) {
				XMoveWindow(dpy, _decks[n].self,
					    x1 * XSTEP, i * XSTEP);
			}
			s -= dx;
			if (s < 0) {
				s += dy;
				x1 += step;
			}
			a++;
		}
	}
	XMoveWindow(dpy, _decks[n].self, x, y);
	XSelectInput(dpy, Table, ExposureMask |
		ButtonPressMask | PointerMotionMask);
}

void
tfree(l)
struct tableau **l;
{
	struct tableau *s, *t;

	s = *l;
	*l = NULL;
	while (s != NULL) {
		t = s;
		s = s->next;
		XtFree((XtPointer) t);
	}
}

void 
(*getGame(name))()
char *name;
{
	char _name[256];
	int i = 0;
	extern void initCan();
	extern void initPyd();
	extern void initTwr();
	extern void initCnr();
	extern void initQad();
	extern void initGlf();
	extern void initCal();
	extern void initKdk();
	extern void initSad();

	while (*name) {
		_name[i++] = tolower(*name);
		name++;
	}
	_name[i] = '\0';

	if      (!strcmp(_name, CANFIELD)) {
		return (initCan);
	}
	else if (!strcmp(_name, PYRAMID)) {
		return (initPyd);
	}
	else if (!strcmp(_name, TOWERS)) {
		return (initTwr);
	}
	else if (!strcmp(_name, CORNERS)) {
		return (initCnr);
	}
	else if (!strcmp(_name, QUADRILLE)) {
		return (initQad);
	}
	else if (!strcmp(_name, GOLF)) {
		return (initGlf);
	}
	else if (!strcmp(_name, CALCULATION)) {
		return (initCal);
	}
	else if (!strcmp(_name, KLONDIKE)) {
		return (initKdk);
	}
	else if (!strcmp(_name, SHUFFLE)) {
		return (initSad);
	}
	else {
		return (NULL);
	}
}

static Pixmap
makeCard(n)
int n;
{
	Pixmap tmp;
	XImage *image;
	static XImage images[VARIETY];

	image = &images[n];
	image->width = card_width;
	image->height = card_height;
	image->xoffset = 0;
	image->format = XYBitmap;
	image->data = card_bits[n];
	image->byte_order = LSBFirst;
	image->bitmap_unit = 16;
	image->bitmap_bit_order = LSBFirst;
	image->bitmap_pad = 16;
	image->bytes_per_line = (card_width + 7) / 8;
	image->depth = 1;
	tmp = XCreatePixmap(dpy, RootWindow(dpy, scr),
		       image->width, image->height, DefaultDepth(dpy, scr));
	if (n >= JOKER) {
		XPutImage(dpy, tmp, n == FRAME ? bgc : ngc,
			image, 0, 0, 0, 0, image->width, image->height);
	}
	else {
		XPutImage(dpy, tmp, iscolor ?
			(REDP(n) ? red_gc : (rvflag ? rgc : ngc)) :
			(rvflag ? ngc : (REDP(n) ? ngc : rgc)),
			image, 0, 0, 0, 0, image->width, image->height);
	}
	return (tmp);
}

void
swap(px, py)
int *px, *py;
{
	int temp;

	temp = *px;
	*px = *py;
	*py = temp;
}

#define CSMAX 128
#define PSMAX 1024

struct position {
	char soup;
	char desp;
	char soun;
	char desn;
};

static struct position pstack[PSMAX];
static int psp;

static struct tableau *cstack[CSMAX];
static int csp;

void
pushCard(c)
struct tableau *c;
{
	if (csp < CSMAX) {
		cstack[csp++] = c;
	}
}

struct tableau *
popCard()
{
	if (csp) {
		return (cstack[--csp]);
	}
	else {
		return (NULL);
	}
}

void
clearStack()
{
	csp = 0;
	psp = 0;
}

int (*_when_hold)();
int (*_when_not_hold)();

static void
holdCard(c, n)
int c;
int n;
{
	if (!(*_when_not_hold)(c, n)) {
		if (psp < PSMAX) {
			pstack[psp].soup = c;
			pstack[psp].soun = n;
		}
		if (!havecard) {
			pstack[psp].desp = -1;
			pstack[psp].desn = -1;
			psp++;
		}
	}
}

static void
releaseCard(c, n)
int c;
int n;
{
	if (!(*_when_hold)(c, n)) {
		if (psp < PSMAX) {
			pstack[psp].desp = c;
			pstack[psp++].desn = n;
		}
	}
}

static void
traceCard(back)
int back;
{
	int i, j;
	int jtmp;
	int mtmp;

	jtmp = jstep;
	jstep = 0x7fff;
	mtmp = psp - back;
	(*_rshook)();
	for (i = 0; i < mtmp; i++) {
		(void) (*_when_not_hold)(pstack[i].soup, pstack[i].soun);
		(void) (*_when_hold)(pstack[i].desp, pstack[i].desn);
		do {
			(*_ckhook)();
		}
		while (autopilot);
	}
	jstep = jtmp;
	psp = mtmp < 0 ? 0 : mtmp;
	havecard = False;
	autopilot = False;
	(*_ckhook)();
}

int
saveData()
{
	int i;
	FILE *fp = fopen(gamefile, "w");

	if (fp != NULL) {
		fprintf(fp, "%d %s\n", randomseed, current);
		for (i = 0; i < psp; i++) {
			fprintf(fp, "%d %d %d %d\n",
				pstack[i].soup,
				pstack[i].desp,
				pstack[i].soun,
				pstack[i].desn);
		}
		fclose(fp);
		return (0);
	}
	return (-1);
}

int
loadData()
{
	char game_name[256];
	FILE *fp = fopen(gamefile, "r");
	void (*getGame())(), (*run)();

	if (fp != NULL) {
		int a, b, c, d, i;
		i = 0;
		fscanf(fp, "%d %s\n", &randomseed, game_name);
		while (fscanf(fp, "%d %d %d %d\n", &a, &b, &c, &d) != EOF) {
			pstack[i].soup = a;
			pstack[i].desp = b;
			pstack[i].soun = c;
			pstack[i].desn = d;
			i++;
		}
		fclose(fp);
		XUnmapWindow(dpy, Table);
		run = getGame(game_name);
		run();
		psp = i;
		traceCard(0);
		XMapWindow(dpy, Table);
		return (0);
	}
	return (-1);
}

void
undo1Step()
{
	if (psp) {
		XUnmapWindow(dpy, Table);
		traceCard(1);
		XMapWindow(dpy, Table);
	}
}
