/*
 * xsols -- Solitaire card games for X11
 *
 * $Header: menu.c,v 0.11 94/12/18 14:33:29 miyazawa Rel $
 */

#ifndef USE_ATHENA

/* with Motif */

#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/PushBG.h>
#include <Xm/CascadeBG.h>
#include <Xm/SeparatoG.h>
#include <Xm/RowColumn.h>
#include <Xm/DrawingA.h>
#include <Xm/MessageB.h>
#include <Xm/DialogS.h>
#include <Xm/Text.h>
#include "modoki.h"

#define SORRY "Sorry, Not Supported yet.\n"
#define HELP "Message"

static Widget end_dialog, help_dialog;
static XmString dead_msg, good_msg;
static String message[GAMEMAX + 1];
static void createEndDialog();
static void createHelpDialog();


String fallbacks[] = {
	"*shadowThickness:	2",
	"*highlightThickness:	0",
	"*fontList:	fixed",
	"*board*dialogStyle:	DIALOG_APPLICATION_MODAL",
	"*end_dialog_popup*mwmDecorations:	0",
	"*end_dialog.okLabelString:	YES",
	"*end_dialog.cancelLabelString:	NO",
	"*end_dialog.again.labelString:	AGAIN",
	"*end_dialog*marginWidth:	10",
	"*end_dialog*marginHeight:	4",
	"*end_dialog*dialogType:	DIALOG_INFORMATION",
	"*file.mnemonic:	F",
	"*file.labelString:	File",
	"*game.mnemonic:	G",
	"*game.labelString:	Game",
	"*help.mnemonic:	H",
	"*help.labelString:	Help",
	"*help_dialog*defaultPosition:	True",
	"*help_dialog*dialogTitle:	HELP",
	"*help_dialog*Spacing:	10",
	"*help_dialog*list.cursorPositionVisible:	False",
	"*help_dialog*list.editMode:	MULTI_LINE_EDIT",
	"*help_dialog*list.editable:	False",
	"*help_dialog*list.scrollHorizontal:	False",
	"*help_dialog*list.scrollVertical:	False",
	"*help_dialog*listSW.width:	100",
	"*help_dialog*listSW.height:	50",
	"*help_dialog*listSW.topAttachment:	ATTACH_WIDGET",
	"*help_dialog*listSW.bottomAttachment:	ATTACH_WIDGET",
	"*help_dialog*listSW.leftAttachment:	ATTACH_WIDGET",
	"*help_dialog*listSW.rightAttachment:	ATTACH_WIDGET",
	"*help_dialog*ok.topAttachment:	ATTACH_NONE",
	"*help_dialog*ok.bottomAttachment:	ATTACH_FORM",
	"*help_dialog*ok.leftAttachment:	ATTACH_POSITION",
	"*help_dialog*ok.rightAttachment:	ATTACH_POSITION",
	"*help_dialog*ok.rightPosition:	70",
	"*help_dialog*ok.leftPosition:	30",
	"*help_dialog*ok.labelString:	Dismiss",
	"*help_dialog*Offset:	10",
	NULL,
};

Widget
createBoard(parent)
Widget parent;
{
	Widget form, board, menubar;
	Widget file, game, help;
	Widget files[6], games[GAMEMAX], helps[2];
	Widget _file, _game, _help, sep;
	Arg args[10];
	int c;
	int i;
	static void Menu();
	static void Help();
	extern void (*getGame())();
	extern char *solitaire[];

	form = XtCreateManagedWidget
		("form", xmFormWidgetClass, parent, NULL, 0);
	board = XtCreateManagedWidget
		("board", xmDrawingAreaWidgetClass, form, NULL, 0);
	menubar = XmCreateMenuBar
		(form, "menubar", NULL, 0);
	XtManageChild(menubar);


	_file = XmCreatePulldownMenu(menubar, "_file", NULL, 0);
	XtSetArg(args[0], XmNsubMenuId, _file);
	file = XmCreateCascadeButtonGadget(menubar, "file", args, 1);
	XtManageChild(file);
	files[5] = XmCreatePushButtonGadget(_file, "undo", NULL, 0);
	files[4] = XmCreatePushButtonGadget(_file, "load", NULL, 0);
	files[3] = XmCreatePushButtonGadget(_file, "save", NULL, 0);
	files[2] = XmCreatePushButtonGadget(_file, "same", NULL, 0);
	files[1] = XmCreatePushButtonGadget(_file, "next", NULL, 0);
	sep = XmCreateSeparatorGadget(_file, "separator", NULL, 0);
	files[0] = XmCreatePushButtonGadget(_file, "quit", NULL, 0);
	XtManageChildren(files, XtNumber(files));
	XtManageChild(sep);

	_game = XmCreatePulldownMenu(menubar, "_game", NULL, 0);
	XtSetArg(args[0], XmNsubMenuId, _game);
	game = XmCreateCascadeButtonGadget(menubar, "game", args, 1);
	XtManageChild(game);

	for (i = 0; i < XtNumber(games); i++) {
		games[i] = XmCreatePushButtonGadget
			(_game, solitaire[i], NULL, 0);
		XtAddCallback(games[i], XmNactivateCallback,
			getGame(solitaire[i]), 0);
	}
	XtManageChildren(games, XtNumber(games));

	_help = XmCreatePulldownMenu(menubar, "_help", NULL, 0);
	XtSetArg(args[0], XmNsubMenuId, _help);
	help = XmCreateCascadeButtonGadget(menubar, "help", args, 1);
	XtManageChild(help);
	helps[0] = XmCreatePushButtonGadget(_help, "misc", NULL, 0);
	helps[1] = XmCreatePushButtonGadget(_help, "rule", NULL, 0);
	XtManageChildren(helps, XtNumber(helps));

	c = 0;
	XtSetArg(args[c], XmNresizePolicy, XmRESIZE_NONE);c++;
	XtSetValues(form, args, c);

	c = 0;
	XtSetArg(args[c], XmNtopAttachment, XmATTACH_FORM);c++;
	XtSetArg(args[c], XmNleftAttachment, XmATTACH_FORM);c++;
	XtSetArg(args[c], XmNrightAttachment, XmATTACH_FORM);c++;
	XtSetArg(args[c], XmNmenuHelpWidget, help);c++;
	XtSetValues(menubar, args, c);

	c = 0;
	XtSetArg(args[c], XmNleftAttachment, XmATTACH_FORM);c++;
	XtSetArg(args[c], XmNrightAttachment, XmATTACH_FORM);c++;
	XtSetArg(args[c], XmNbottomAttachment, XmATTACH_FORM);c++;
	XtSetArg(args[c], XmNtopAttachment, XmATTACH_WIDGET);c++;
	XtSetArg(args[c], XmNtopWidget, menubar);c++;
	XtSetArg(args[c], XmNresizePolicy, XmRESIZE_NONE);c++;
	XtSetValues(board, args, c);

	XtAddCallback(files[0], XmNactivateCallback, (XtCallbackProc) Menu, (XtPointer) 0);/* bye */
	XtAddCallback(files[1], XmNactivateCallback, (XtCallbackProc) Menu, (XtPointer) 1);/* restart */
	XtAddCallback(files[2], XmNactivateCallback, (XtCallbackProc) Menu, (XtPointer) 2);/* retry */
	XtAddCallback(files[3], XmNactivateCallback, (XtCallbackProc) Menu, (XtPointer) 3);/* save */
	XtAddCallback(files[4], XmNactivateCallback, (XtCallbackProc) Menu, (XtPointer) 4);/* load */
	XtAddCallback(files[5], XmNactivateCallback, (XtCallbackProc) Menu, (XtPointer) 5);/* undo */

	XtAddCallback(helps[0], XmNactivateCallback, (XtCallbackProc) Help, (XtPointer) 0);
	XtAddCallback(helps[1], XmNactivateCallback, (XtCallbackProc) Help, (XtPointer) 1);

	createEndDialog(board);
	createHelpDialog(board);

	return (board);
}

static void
createHelpDialog(parent)
Widget parent;
{
	Arg args[10];
	int c;
	Widget ok, sw, list;
	static XtResource resources[] = {
		{CANFIELD, HELP, XtRString, sizeof(String),
		(Cardinal) &message[0], XtRString, SORRY},
		{PYRAMID, HELP, XtRString, sizeof(String),
		(Cardinal) &message[1], XtRString, SORRY},
		{TOWERS, HELP, XtRString, sizeof(String),
		(Cardinal) &message[2], XtRString, SORRY},
		{CORNERS, HELP, XtRString, sizeof(String),
		(Cardinal) &message[3], XtRString, SORRY},
		{QUADRILLE, HELP, XtRString, sizeof(String),
		(Cardinal) &message[4], XtRString, SORRY},
		{GOLF, HELP, XtRString, sizeof(String),
		(Cardinal) &message[5], XtRString, SORRY},
		{CALCULATION, HELP, XtRString, sizeof(String),
		(Cardinal) &message[6], XtRString, SORRY},
		{KLONDIKE, HELP, XtRString, sizeof(String),
		(Cardinal) &message[7], XtRString, SORRY},
		{SHUFFLE, HELP, XtRString, sizeof(String),
		(Cardinal) &message[8], XtRString, SORRY},
		{"misc", "Misc", XtRString, sizeof(String),
		(Cardinal) &message[GAMEMAX], XtRString, SORRY},
	};

	help_dialog = XmCreateFormDialog(parent, "help_dialog", NULL, 0);
	list = XmCreateScrolledText(help_dialog, "list", NULL, 0);
	XtManageChild(list);
	sw = XtParent(list);
	ok = XtCreateManagedWidget
		("ok", xmPushButtonGadgetClass, help_dialog, NULL, 0);

	c = 0;
	XtSetArg(args[c], XmNbottomWidget, ok);c++;
	XtSetValues(sw, args, c);

	c = 0;
	XtSetArg(args[c], XmNdefaultButton, ok);c++;
	XtSetValues(help_dialog, args, c);

	XtGetApplicationResources(list, 0, resources,
		XtNumber(resources), NULL, (Cardinal) 0);

}

static void
createEndDialog(parent)
Widget parent;
{
	static XtResource resources[] = {
		{"deadMessage", "DeadMessage", XmRXmString, sizeof(XmString),
		(Cardinal) &dead_msg, XtRString, "You dead. New game?"},
		{"goodMessage", "GoodMessage", XmRXmString, sizeof(XmString),
		(Cardinal) &good_msg, XtRString, "You win. New game?"},
	};
	Widget yes, again, w;
	static void Menu();

	end_dialog = XmCreateQuestionDialog(parent, "end_dialog", NULL, 0);
	XtGetApplicationResources(end_dialog, 0, resources,
		XtNumber(resources), NULL, (Cardinal) 0);

	w = XmMessageBoxGetChild(end_dialog, XmDIALOG_HELP_BUTTON);
	XtUnmanageChild(w);
	yes = XmMessageBoxGetChild(end_dialog, XmDIALOG_DEFAULT_BUTTON);
	XtAddCallback(yes, XmNactivateCallback, (XtCallbackProc) Menu, (XtPointer) 1);

#if XmREVISION > 1
	again = XmCreatePushButtonGadget(end_dialog, "again", NULL, 0);
	XtManageChild(again);
	XtAddCallback(again, XmNactivateCallback, (XtCallbackProc) Menu, (XtPointer) 2);
#endif
}

static void
Menu(w, c)
Widget w;
int c;
{
	extern void (*_rshook)();
	extern void undo1Step();

	switch(c) {
	case 0:
		exit(0);
		break;
	case 1:
		randomseed++;
	case 2:
		(*_rshook)();
		(*_ckhook)();
		break;
	case 3:
		saveData();
		break;
	case 4:
		loadData();
		break;
	case 5:
		undo1Step();
		break;
	default:
		break;
	}
}

static void
Help(w, c)
Widget w;
int c;
{
	Widget t;
	String s;

	if (!c) {
		s = message[GAMEMAX];
	}
	else if (!strcmp(current, CANFIELD)) {
		s = message[0];
	}
	else if (!strcmp(current, PYRAMID)) {
		s = message[1];
	}
	else if (!strcmp(current, TOWERS)) {
		s = message[2];
	}
	else if (!strcmp(current, CORNERS)) {
		s = message[3];
	}
	else if (!strcmp(current, QUADRILLE)) {
		s = message[4];
	}
	else if (!strcmp(current, GOLF)) {
		s = message[5];
	}
	else if (!strcmp(current, CALCULATION)) {
		s = message[6];
	}
	else if (!strcmp(current, KLONDIKE)) {
		s = message[7];
	}
	else if (!strcmp(current, SHUFFLE)) {
		s = message[8];
	}
	else {
		s = "HOGEHOGE\n";
	}
	t = XtNameToWidget(help_dialog, "listSW.list");
	XmTextSetString(t, s);
	XtManageChild(help_dialog);
}

static void
bluePopdown()
{
	XtUnmanageChild(end_dialog);
}

void
bluePopup(status) 
int status;
{
	Arg arg;

	if (status < 0) {
		XtSetArg(arg, XmNmessageString, dead_msg);
	}
	else if (status > 0){
		XtSetArg(arg, XmNmessageString, good_msg);
	}
	else {
		return;
	}
	XtSetValues(end_dialog, &arg, 1);
	XtManageChild(end_dialog);
}

#else /* USE_ATHENA */

/* with Athena */

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Viewport.h>

#include "modoki.h"

#define SORRY "Sorry, Not Supported yet.\n"
#define HELP "Message"

#define min(x,y) (((x)<(y))?(x):(y))
#define max(x,y) (((x)>(y))?(x):(y))

static Widget end_dialog, help_dialog;
static String dead_msg, good_msg;
static String message[GAMEMAX + 1];
static void createEndDialog();
static void createHelpDialog();
static void bluePopdown();

String fallbacks[] = {
	"*whiteColor:#efe",
	"*redColor:#f44",
	"*background:#aba",
	"*internalWidth: 3",
	"*menubar.borderWidth:0",
	"*menubar.file.borderWidth:0",
	"*menubar.game.borderWidth:0",
	"*menubar.help.borderWidth:0",
	"*menubar.orientation:horizontal",
	"*end_dialog*borderWidth:1",
	"*help_dialog*borderWidth:1",
	"*help_dialog*sw.allowVert:True",
	"*help_dialog*sw.useRight:True",
	"*help_dialog*sw.height:200",
	"*help_dialog*sw.width:540",
	"*help_dialog*list.misc:\
****************************************************************************\\n\
\\n\
 xsols -- Solitaire card games for X11\\n\
\\n\
 Copyright (c) 1992  Advanced Integration Technology, Inc.\\n\
\\n\
 Permission to use, copy, modify, and distribute this software and its\\n\
 documentation for any purpose and without fee is hereby granted, provided\\n\
 that the above copyright notice appear in all copies and that both that\\n\
 copyright notice and this permission notice appear in supporting\\n\
 documentation, and that the name of AIT not be used in advertising or\\n\
 publicity pertaining to distribution of the software without specific,\\n\
 written prior permission.  AIT makes no representations about the\\n\
 suitability of this software for any purpose.  It is provided \"as is\"\\n\
 without express or implied warranty.\\n\
\\n\
 Author:  Kunio Miyazawa\\n\
\\n\
****************************************************************************\\n",
	"*help_dialog*list.calculation:under construction",
	"*help_dialog*list.pyramid:under construction",
	"*help_dialog*list.klondike:under construction",
	"*help_dialog*list.quadrille:under construction",
	"*help_dialog*list.golf:under construction",
	"*help_dialog*list.shaffle:under construction",
	"*help_dialog*list.corners:under construction",
	"*help_dialog*list.towers:under construction",
	"*help_dialog*list.canfield:under construction",
	"*help_dialog*list.shuffle:under construction",
	NULL
};

Widget
createBoard(parent)
Widget parent;
{
	Widget form, board, menubar;
	Widget file, game, help;
	Widget files[3], games[GAMEMAX], helps[2];
	Widget _file, _game, _help, sep;
	Arg args[10];
	int c;
	int i;
	extern void Menu();
	extern void Help();
	extern void (*getGame())();
	extern char *solitaire[];

	form = XtCreateManagedWidget
		("form", boxWidgetClass, parent, NULL, 0);
	menubar = XtCreateManagedWidget
		("menubar", boxWidgetClass, form, NULL, 0);
	board = XtCreateManagedWidget
		("board", boxWidgetClass, form, NULL, 0);

	XtSetArg(args[0], XtNmenuName, "_file");
	file = XtCreateManagedWidget
		("file", menuButtonWidgetClass, menubar, args, 1);
	_file = XtCreatePopupShell
		("_file", simpleMenuWidgetClass, file, NULL, 0);
	files[5] = XtCreateManagedWidget
		("undo", smeBSBObjectClass, _file, NULL, 0);
	files[4] = XtCreateManagedWidget
		("load", smeBSBObjectClass, _file, NULL, 0);
	files[3] = XtCreateManagedWidget
		("save", smeBSBObjectClass, _file, NULL, 0);
	files[2] = XtCreateManagedWidget
		("same", smeBSBObjectClass, _file, NULL, 0);
	files[1] = XtCreateManagedWidget
		("next", smeBSBObjectClass, _file, NULL, 0);
	files[0] = XtCreateManagedWidget
		("quit", smeBSBObjectClass, _file, NULL, 0);
	XtAddCallback(files[0], XtNcallback, Menu, (XtPointer) 0);/* bye */
	XtAddCallback(files[1], XtNcallback, Menu, (XtPointer) 1);/* restart */
	XtAddCallback(files[2], XtNcallback, Menu, (XtPointer) 2);/* retry */
	XtAddCallback(files[3], XtNcallback, Menu, (XtPointer) 3);/* save */
	XtAddCallback(files[4], XtNcallback, Menu, (XtPointer) 4);/* load */
	XtAddCallback(files[5], XtNcallback, Menu, (XtPointer) 5);/* undo */

	XtSetArg(args[0], XtNmenuName, "_game");
	game = XtCreateManagedWidget
		("game", menuButtonWidgetClass, menubar, args, 1);
	_game = XtCreatePopupShell
		("_game", simpleMenuWidgetClass, game, NULL, 0);
	for (i = 0; i < XtNumber(games); i++) {
		games[i] = XtCreateManagedWidget
			(solitaire[i], smeBSBObjectClass, _game, NULL, 0);
		XtAddCallback(games[i], XtNcallback,
			getGame(solitaire[i]), (XtPointer) 0);
	}

	XtSetArg(args[0], XtNmenuName, "_help");
	help = XtCreateManagedWidget
		("help", menuButtonWidgetClass, menubar, args, 1);
	_help = XtCreatePopupShell
		("_help", simpleMenuWidgetClass, help, NULL, 0);
	helps[0] = XtCreateManagedWidget
		("misc", smeBSBObjectClass, _help, NULL, 0);
	helps[1] = XtCreateManagedWidget
		("rule", smeBSBObjectClass, _help, NULL, 0);
	XtAddCallback(helps[0], XtNcallback, Help, (XtPointer) 0);
	XtAddCallback(helps[1], XtNcallback, Help, (XtPointer) 1);

	createEndDialog(board);
	createHelpDialog(board);

	return (board);
}

static void
createHelpDialog(parent)
Widget parent;
{
	Arg args[10];
	int c;
	Widget form, ok, sw, list;
	void closeHelp();
	static XtResource resources[] = {
		{CANFIELD, HELP, XtRString, sizeof(String),
		(Cardinal) &message[0], XtRString, SORRY},
		{PYRAMID, HELP, XtRString, sizeof(String),
		(Cardinal) &message[1], XtRString, SORRY},
		{TOWERS, HELP, XtRString, sizeof(String),
		(Cardinal) &message[2], XtRString, SORRY},
		{CORNERS, HELP, XtRString, sizeof(String),
		(Cardinal) &message[3], XtRString, SORRY},
		{QUADRILLE, HELP, XtRString, sizeof(String),
		(Cardinal) &message[4], XtRString, SORRY},
		{GOLF, HELP, XtRString, sizeof(String),
		(Cardinal) &message[5], XtRString, SORRY},
		{CALCULATION, HELP, XtRString, sizeof(String),
		(Cardinal) &message[6], XtRString, SORRY},
		{KLONDIKE, HELP, XtRString, sizeof(String),
		(Cardinal) &message[7], XtRString, SORRY},
		{SHUFFLE, HELP, XtRString, sizeof(String),
		(Cardinal) &message[8], XtRString, SORRY},
		{"misc", "Misc", XtRString, sizeof(String),
		(Cardinal) &message[GAMEMAX], XtRString, SORRY},
	};

	help_dialog = XtCreatePopupShell
		("help_dialog", transientShellWidgetClass, parent, NULL, 0);
	form = XtCreateManagedWidget
		("form", formWidgetClass, help_dialog, NULL, 0);
	sw = XtCreateManagedWidget
		("sw", viewportWidgetClass, form, NULL, 0);
	list = XtCreateManagedWidget
		("list", labelWidgetClass, sw, NULL, 0);

	XtSetArg(args[0], XtNfromVert, sw);
	ok = XtCreateManagedWidget
		("ok", commandWidgetClass, form, args, 1);
	XtAddCallback(ok, XtNcallback, closeHelp, 0);

	XtGetApplicationResources(list, 0, resources,
		XtNumber(resources), NULL, (Cardinal) 0);
}

static void
createEndDialog(parent)
Widget parent;
{
	static XtResource resources[] = {
		{"deadMessage", "DeadMessage", XtRString, sizeof(String),
		(Cardinal) &dead_msg, XtRString, "You dead. Try again?"},
		{"goodMessage", "GoodMessage", XtRString, sizeof(String),
		(Cardinal) &good_msg, XtRString, "You win. Try again?"},
	};
	Widget yes, w;
	extern void Menu();

	end_dialog = XtCreatePopupShell
		("end_dialog", transientShellWidgetClass, parent, NULL, 0);
	XtGetApplicationResources(end_dialog, 0, resources,
		XtNumber(resources), NULL, (Cardinal) 0);

	w = XtCreateManagedWidget
		("dialog", dialogWidgetClass, end_dialog, NULL, 0);
	XawDialogAddButton(w, "yes", Menu, (XtPointer) 2);
	XawDialogAddButton(w, "no", bluePopdown, (XtPointer) 0);
}

static void
Menu(w, c)
Widget w;
int c;
{
	extern void (*_rshook)();
	extern void undo1Step();

	switch(c) {
	case 0:
		exit(0);
		break;
	case 1:
		randomseed++;
		(*_rshook)();
		(*_ckhook)();
		break;
	case 2:
		bluePopdown();
		(*_rshook)();
		(*_ckhook)();
		break;
	case 3:
		saveData();
		break;
	case 4:
		loadData();
		break;
	case 5:
		undo1Step();
		break;
	default:
		break;
	}
}

static void
closeHelp()
{
	XtPopdown(help_dialog);
}

static void
Help(w, c)
Widget w;
int c;
{
	Position popup_x, popup_y, top_x, top_y;
	Dimension width, height;
	Dimension popup_width, popup_height;
	Dimension top_width, top_height, border_width;
	Widget top;
	Widget t;
	String s;
	int n;
	Arg args[4];

	if (!c) {
		s = message[GAMEMAX];
	}
	else if (!strcmp(current, CANFIELD)) {
		s = message[0];
	}
	else if (!strcmp(current, PYRAMID)) {
		s = message[1];
	}
	else if (!strcmp(current, TOWERS)) {
		s = message[2];
	}
	else if (!strcmp(current, CORNERS)) {
		s = message[3];
	}
	else if (!strcmp(current, QUADRILLE)) {
		s = message[4];
	}
	else if (!strcmp(current, GOLF)) {
		s = message[5];
	}
	else if (!strcmp(current, CALCULATION)) {
		s = message[6];
	}
	else if (!strcmp(current, KLONDIKE)) {
		s = message[7];
	}
	else if (!strcmp(current, SHUFFLE)) {
		s = message[8];
	}
	else {
		return;
	}

	t = XtNameToWidget(help_dialog, "form.sw.list");

	n = 0;
	XtSetArg(args[n], XtNlabel, s);n++;
	XtSetValues(t, args, n);

	XtRealizeWidget(help_dialog);
	for (top = help_dialog; XtParent(top) != NULL; top = XtParent(top));

	n = 0;
	XtSetArg(args[n], XtNx, &top_x); n++;
	XtSetArg(args[n], XtNy, &top_y); n++;
	XtSetArg(args[n], XtNwidth, &top_width); n++;
	XtSetArg(args[n], XtNheight, &top_height); n++;
	XtGetValues(top, args, n);

	n = 0;
	XtSetArg(args[n], XtNwidth, &popup_width); n++;
	XtSetArg(args[n], XtNheight, &popup_height); n++;
	XtSetArg(args[n], XtNborderWidth, &border_width); n++;
	XtGetValues(help_dialog, args, n);

	popup_x = max(0,
		min(top_x + (Position) (top_width - popup_width) / 2, 
		DisplayWidth(dpy, scr) -
		(Position) popup_width - 2 * (Position) border_width));
	popup_y = max(0, 
		min(top_y + (Position) (top_height - popup_height) / 2,
		DisplayHeight(dpy, scr) -
		(Position) popup_height - 2 * (Position) border_width));
	n = 0;
	XtSetArg(args[n], XtNx, popup_x); n++;
	XtSetArg(args[n], XtNy, popup_y); n++;
	XtSetValues(help_dialog, args, n);

	XtPopup(help_dialog, XtGrabExclusive);

	XWarpPointer(dpy,
	       None,
	       XtWindow(help_dialog), 
	       0, 0, top_width, top_height,
	       popup_width / 2, popup_height / 2);
}

static void
bluePopdown()
{
	XtPopdown(end_dialog);
}

void
bluePopup(status) 
int status;
{
	Position popup_x, popup_y, top_x, top_y;
	Dimension popup_width, popup_height;
	Dimension top_width, top_height, border_width;
	Widget top;
	int n;
	Arg args[4];

	n = 0;
	if (status < 0) {
		XtSetArg(args[n], XtNlabel, dead_msg);n++;
	}
	else if (status > 0) {
		XtSetArg(args[n], XtNlabel, good_msg);n++;
	}
	else {
		return;
	}
	XtSetValues(XtNameToWidget(end_dialog, "dialog"), args, n);

	XtRealizeWidget(end_dialog);
	for (top = end_dialog; XtParent(top) != NULL; top = XtParent(top));

	n = 0;
	XtSetArg(args[n], XtNx, &top_x); n++;
	XtSetArg(args[n], XtNy, &top_y); n++;
	XtSetArg(args[n], XtNwidth, &top_width); n++;
	XtSetArg(args[n], XtNheight, &top_height); n++;
	XtGetValues(top, args, n);

	n = 0;
	XtSetArg(args[n], XtNwidth, &popup_width); n++;
	XtSetArg(args[n], XtNheight, &popup_height); n++;
	XtSetArg(args[n], XtNborderWidth, &border_width); n++;
	XtGetValues(end_dialog, args, n);

	popup_x = max(0,
		min(top_x + (top_width - popup_width) / 2, 
		DisplayWidth(dpy, scr) -
		popup_width - 2 * border_width));
	popup_y = max(0, 
		min(top_y + (top_height - popup_height) / 2,
		DisplayHeight(dpy, scr) -
		popup_height - 2 * border_width));
	n = 0;
	XtSetArg(args[n], XtNx, popup_x); n++;
	XtSetArg(args[n], XtNy, popup_y); n++;
	XtSetValues(end_dialog, args, n);

	XtPopup(end_dialog, XtGrabExclusive);

	XWarpPointer(dpy,
	       None,
	       XtWindow(end_dialog), 
	       0, 0, top_width, top_height,
	       popup_width / 2, popup_height / 2);
}
#endif /* USE_ATHENA */
