/*
 * quadrille !!
 *
 * $Header: qad.c,v 0.7 94/10/31 15:17:28 miyazawa Rel $
 */

#include "modoki.h"

#define _MAX 40
#define TOP_X 16
#define TOP_Y 40

#define COLORS(a) SUIT(a)

static int card[_MAX];
static int sp;
static int found[8];
static int holdcard;
static int from;
static struct tableau *tableau[3];
static int px[12], py[12];
static int whenHold();
static int whenNotHold();


void
initQad()
{
	static int check();
	static void restart();

	_ckhook = check;
	_rshook = restart;
	_when_hold = whenHold;
	_when_not_hold = whenNotHold;
	current = QUADRILLE;
	(*_rshook)();
	(*_ckhook)();
}

static void
restart()
{
	register i, j;
	static long seed;
	
	srandom(randomseed);
	XUnmapSubwindows(dpy, Table);
	clearStack();

	/*
	  0 1 2 3    5 6 7
                   4       8
                    1110 9
	*/
		 
	px[0] = TOP_X;
	py[0] = TOP_Y;

	for (i = 0; i < 3; i++) {
		px[i + 1] = TOP_X + 50 * (i + 1);
		py[i + 1] = py[0];
	}
	px[4] = px[3] + 70;
	px[8] = px[4] + 170;
	py[4] = py[8] = 124 + TOP_Y;

	px[5] = px[11] = px[4] + 35;
	px[6] = px[10] = px[5] + 50;
	px[7] = px[9] = px[6] + 50;
	py[5] = py[6] = py[7] = py[0];
	py[9] = py[10] = py[11] = 244 + TOP_Y;

	for (i = 0; i < 3; i++) {
		tfree(&tableau[i]);
	}

	for (j = 0, i = 0; i < 52; i++) {
		if (i % 13 == 4 || i % 13 == 5 || i % 13 == 11) {
			continue;
		}
		card[j++] = i;
	}
	for (i = 0; i < _MAX; i++) {
		swap(&card[random() % _MAX], &card[i]);
	}

	for (i = 0; i < 8; i++) {
		found[i] = (3 - i % 4) * 13 + 4 + i / 4;
	}

	for (i = 1; i < 4; i++) {
		PUTCARD(FRAME, i);
	}
	for (i = 0; i < 4; i++) {
		putCard(px[4] + 50 * ((i + 1) % 2) + 60,
			py[0] + 80 * (i / 2) + 80,
			11 + i * 13, -1);
	}
	for (i = 0; i < 8; i++) {
		PUTCARD(found[i], i + 4);
	}
	PUTCARD(card[0], 0);

	sp = 0;
	holdcard = -1;
	havecard = False;
	from = 0;
	_status = 0;
}

static int
check()
{
	register i, j, k;
	int sou, des;
	int f;

	autopilot = False;
	sou = des = 0;

	for (i = 0; i < 8; i++){
		if (NUMB(found[i]) != (i > 3 ? JACK : KING)) {
			goto skip;
		}
	}
	return (1);
skip:
	for (i = 0; i < 3; i++) {
		if (tableau[i] != NULL) {
			f = tableau[i]->n;
			for (j = 0; j < 8; j++) {
				if (COLORS(f) == COLORS(found[j])) {
					if ((NUMB(found[j]) == NUMB(f + 1) &&
						j < 4) ||
					    (NUMB(found[j] + 1) == NUMB(f) &&
						j > 3)) {
						sou = i + 1;
						des = j + 4;
						if (sp >= _MAX) {
							autopilot = True;
						}
						goto ok;
					}
				}
			}
		}
	}
	if (sp < _MAX) {
		f = card[sp];
		for (j = 0; j < 8; j++) {
			if (COLORS(f) == COLORS(found[j])) {
				if ((NUMB(found[j]) == NUMB(f + 1) && j < 4) ||
				    (NUMB(found[j] + 1) == NUMB(f) && j > 3)) {
					sou = 0;
					des = j + 4;
					goto ok;
				}
			}
		}
	}
	else {
		return (-1);
	}
ok:
	if (autopilot) {
		if (havecard) {
			jumpMove(px[des], py[des], holdcard);
			whenHold(des);
			havecard = False;
		}
		else {
			whenNotHold (sou);
		}
	}
	return (0);
}

static int
whenNotHold(pos)
int pos;
{
	if (pos < 4 && pos > 0) {
		if (tableau[pos - 1] != NULL) {
			holdcard = tableau[pos - 1]->n;
			havecard = True;
			from = pos;
			return (0);
		}
	}
	else if (pos == 0) {
		holdcard = card[sp];
		havecard = True;
		from = pos;
		return (0);
	}
	return (-1);
}

static int
whenHold(pos)
int pos;
{
	struct tableau *temp;
	
	if (pos == from || pos < 0) {
		holdcard = -1;
		return (-1);
	}
	if (pos > 3) {
		if (COLORS(holdcard) == COLORS(found[pos - 4])) {
			if ((NUMB(found[pos - 4]) ==
			     NUMB(holdcard + 1) && pos < 8) ||
			    (NUMB(found[pos - 4] + 1) ==
			     NUMB(holdcard) && pos > 7)) {
				found[pos - 4] = holdcard;
				PUTCARD(holdcard, pos);
				if (from) {
					temp = tableau[from - 1];
					tableau[from - 1] = temp->next;
					if (temp->next != NULL) {
						py[from] = temp->next->tmp;
						PUTCARD(temp->next->n, from);
					}
					else {
						py[from] = TOP_Y;
					}
					XtFree((XtPointer) temp);
				}
				else {
					if (sp < _MAX - 1) {
						PUTCARD(card[++sp], 0);
					}
					else {
						sp = _MAX;
					}
				}
				holdcard = -1;
				return (0);
			}
		}
	}
	else if (pos > 0 && pos < 4 && from == 0 && py[pos] < 305) {
		if (tableau[pos - 1] != NULL) {
			py[pos] += 15;
		}
		temp = talloc();
		temp->next = tableau[pos - 1];
		temp->tmp = py[pos];
		temp->n = holdcard;
		tableau[pos - 1] = temp;
		PUTCARD(holdcard, pos);
		holdcard = -1;
		if (sp < _MAX - 1) {
			PUTCARD(card[++sp], 0);
		}
		else {
			sp = _MAX;
		}
		holdcard = -1;
		return (0);
	}
	holdcard = -1;
	return (-1);
}
